/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.modelling;


import java.util.ArrayList;

import javax.swing.Action;
import javax.swing.JMenu;
import javax.swing.JMenuItem;

/**
 * This class serves an abstraction from a cascading menu that only supports
 * actions (no checkboxes or radiobuttons). This means it should be easier
 * and more flexible to implement for example in a virtual reality environment
 * where no elaborate widget set is available.
 * 
 * @author Ulrich Pinkall
 */
public class ActionTree {
	
	private Action action;
	private String name;
	private ArrayList<ActionTree> items = new ArrayList<ActionTree>();

	/**
	 * Creates an action tree as a toplevel list of other action trees. Should be used
	 * only as the tree root.
	 */
	public ActionTree() {
		items = new ArrayList<ActionTree>();
	}
	
	/**
	 * Creates an action tree to be used for a cascading submenu.
	 * @param name The string to be presented on the menu item that leads
	 * to the submenu. Tree nodes are always action trees themselves. They
	 * are considered as submenues if their action is null. Otherwise they
	 * are expected to have a non-null action.
	 */
	public ActionTree(String name) {
		this.name = name;
		items = new ArrayList<ActionTree>();
	}
	
	/**
	 * Creates a leaf for an action tree.
	 * @param a
	 */
	public ActionTree(Action a) {
		action = a;
	}
	
	/**
	 * Returns the action of this tree node. If <code>null</code>, the node
	 * is considered as a submenu, to be accessed via <code>getItem(int)</code>.
	 */
	public Action getAction() {
		return action;
	}
	
	/**
	 * Returns the number of items in the submenu represented by this action
	 * tree. Supposed to return zero if the action of this tree is non-null.
	 */
	public int getItemCount() {
		return items.size();
	}
	
	/**
	 * Returns the item with index <code>j</code> in this submenu.
	 */
	public ActionTree getItem(int j) {
		return items.get(j);
	}
	
	/**
	 * Adds an action tree (representing either a submenu or an action) to
	 * this menu. Throws an IllegalStateException if the action of this
	 * tree is non-null.
	 */
	public void add(ActionTree t) {
		if (action == null) {
			items.add(t);
		} else {
			throw new IllegalStateException();
		}
	}
	
	/**
	 * Adds an action (encapsuled by an action tree) to this menu. Throws an
	 * IllegalStateException if the action of this tree is non-null.
	 */
	public void add(Action a) {
		if (action == null) { 
			add(new ActionTree(a));
		} else {
			throw new IllegalStateException();
		}
	}
	
	/**
	 * Removes an action tree that is a direct child of this tree node.
	 */
	public void remove(ActionTree t) {
		items.remove(t);
	}
	
	/**
	 * Removes all action trees that are a direct children of this tree node.
	 */
	public void removeAll() {
		items.clear();
	}
	
	/**
	 * Returns a <code>JMenuItem</code>, which is either an action button
	 * (if the action is non-null) or a cascading JMenu representing the whole
	 * tree, including all descendents.
	 */
	public JMenuItem asMenuItem() {
		if (action != null) {
			return new JMenuItem(action);
		} else {
			JMenu menu = new JMenu(name);
			for (ActionTree t : items) {
				menu.add(t.asMenuItem());
			}
			return menu;
		}
	}

	/**
	 * Returns the name of this submenu. Supposed to be <code>null</code> if
	 * the action is non-null.
	 */
	public String getName() {
		return name;
	}

	public void setEnabled(boolean b) {
		if (action != null) action.setEnabled(b);
		for (ActionTree item : items) item.setEnabled(b);
	}
}
