/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.modelling;


import java.awt.event.ActionEvent;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.converters.reflection.PureJavaReflectionProvider;



public class Modeller {
	private Inspector inspector;
	private ModelContainer modelContainer;
	private PersistencePeer persistencePeer;
    private Object controlScene;
    private Object viewScene;
	private ModellingViewer viewer;
	private boolean updatingViewFromEnvironment = true;
	private HashSet<ModellingTool> tools = new HashSet<ModellingTool>();
	// for serialization
	HashMap<ModellingTool, String> toolToTemplate =
		new HashMap<ModellingTool, String>();
	HashMap<String, ModellingTool> templateNameToTemplate =
		new HashMap<String, ModellingTool>();
	private ModellingTool clipBoardTool;
	private String clipBoardTemplate;
	private Action pasteAction = new AbstractAction("paste") {
    	final static long serialVersionUID = 1984;
    	public void actionPerformed(ActionEvent e) {
    		paste();
    	}
    };
    private final ChangeListener containerListener = new ChangeListener() {
		public void stateChanged(ChangeEvent e) {
			if (updatingViewFromEnvironment) {
				viewer.render();
			}
		}
    };
    private final ChangeListener inspectorListener = new ChangeListener() {
		public void stateChanged(ChangeEvent e) {
			Object o = e.getSource();
			if (modelToTool.containsKey(o)) {
				ModellingTool tool = modelToTool.get(o);
				tool.stateChanged(e);
			}
			Modeller.this.viewer.render();
		}
	};
    private HashMap<String,Object>templateToScene =
    	new HashMap<String,Object>();
    
	private HashMap<Object,ModellingTool> modelToTool =
		new HashMap<Object,ModellingTool>();
	
    private HashMap<String,Integer> templateToCount =
    	new HashMap<String,Integer>();
    
    private LinkedHashMap<Object,Set<String>> objectToExcludedProperties =
    	new LinkedHashMap<Object,Set<String>>();
    
    private ActionTree viewerMenu = new ActionTree();
    private ActionTree inspectMenu = new ActionTree("Inspect");
    private ActionTree sceneInspectMenu = new ActionTree("scene");
    private ActionTree newMenu = new ActionTree("new");
    private ActionTree fileMenu = new ActionTree("File");
    private ActionTree editMenu = new ActionTree("Edit");
    private boolean copyEnabled = false;

    private final ChangeListener toolListener = new ChangeListener() {
		public void stateChanged(ChangeEvent event) {
			ModellingTool tool = (ModellingTool)event.getSource();
			modelContainer.modelChanged(tool.getModel(),tool.getModelType());
		}
    };
    
    /**
     * Creates a modeller with all the needed global data specified. Specific
     * modelling tool templates and objects for the inspect menu are added later.
     * The creation process is to be finished by calling <code>initializeViewer</code>.
     * 
     * @param modelContainer The model container to which all models will be added.
     * @param modelContainerName The name of the model container as it will show
     * in the inspect menu.
     * @param viewer The viewer to which all graphical scenes will be added.
     * @param inspector The facility that allows inspection and editing of objects.
     * @param controlScene The scene to which all control elements will be added.
     * @param excludedPropertiesForControlScene The list of names of properties that
     * should not be exposed when inspecting the control scene.
     */
	public Modeller(
			final ModelContainer modelContainer,
			final ModellingViewer viewer,
			final Inspector inspector,
			final PersistencePeer persistencePeer,
			final Object controlScene,
			final Object viewScene
	) {
		setModelContainer(modelContainer);
		this.viewer = viewer;
		this.inspector = inspector;
		inspector.addChangeListener(inspectorListener);
		this.persistencePeer = persistencePeer;
		this.controlScene = controlScene;
		
		this.viewScene = viewScene;
		viewer.addScene(viewScene);
		viewer.addScene(controlScene);
		
        sceneInspectMenu.add(new AbstractAction("controls") {
        	final static long serialVersionUID = 1984;
        	
        	public void actionPerformed(ActionEvent e) {
        		Modeller.this.inspector.inspect(
        				getControlScene(),
        				objectToExcludedProperties.get(controlScene)
        		);
        	}
        });
        
        sceneInspectMenu.add(new AbstractAction("all tools") {
        	final static long serialVersionUID = 1984;
        	
        	public void actionPerformed(ActionEvent e) {
        		Modeller.this.inspector.inspect(
        				getViewScene(),
        				objectToExcludedProperties.get(viewScene)
        		);
        	}
        });
        
        inspectMenu.add(sceneInspectMenu);
        
        fileMenu.add(new AbstractAction("open session") {
			private static final long serialVersionUID = 8093070908354359952L;

			public void actionPerformed(ActionEvent e) {
				open();
			}
        });
        
        fileMenu.add(new AbstractAction("load scene") {
        	
			private static final long serialVersionUID = 8093070908354359952L;

			public void actionPerformed(ActionEvent e) {
				load();
			}
        });
        fileMenu.add(new AbstractAction("save session") {
        	
			private static final long serialVersionUID = 8093070908354359952L;
			
			public void actionPerformed(ActionEvent e) {
				save();
			}
        });
        editMenu.add(new AbstractAction("store model states") {
        	
        	private static final long serialVersionUID = 8093070908354359952L;
        	
        	public void actionPerformed(ActionEvent e) {
        		storeModelStates();
        	}
        });
        editMenu.add(new AbstractAction("recall model states") {
        	
        	private static final long serialVersionUID = 8093070908354359952L;
        	
        	public void actionPerformed(ActionEvent e) {
        		recallModelStates();
        	}
        });
        editMenu.add(new AbstractAction("clear all") {
        	
        	private static final long serialVersionUID = 8093070908354359952L;
        	
        	public void actionPerformed(ActionEvent e) {
				clear();
			}
        });
        viewerMenu.add(newMenu);
        pasteAction.setEnabled(false);
        viewerMenu.add(pasteAction);
	}
	
	public void clear() {
		HashSet<ModellingTool> tmp = new HashSet<ModellingTool>();
		tmp.addAll(tools);
		for (ModellingTool tool : tmp) {
			removeTool(tool);
		}
	}

	public void load() {
		open(false);
	}
	
	public void open() {
		open(true);
	}
	
	public void open(boolean restoreAll) {
		//JFileChooser chooser = ModellerIO.createFileChooser(null);
		File f = persistencePeer.openFileDialog();
		open(f, restoreAll);
	}
	
	public void open(File f, boolean restoreAll) {
		if (f != null && f.exists()) {
			if (restoreAll) {
				clear();
				templateNameToTemplate.clear();
				templateToCount.clear();
				templateToScene.clear();
			}
			XStream xstream = createXStream();
			try {
				Reader xml = new FileReader(f);
				HashMap in = (HashMap)xstream.fromXML(xml);
				
				if (restoreAll) {
					Object newControlScene = in.get("controlScene");
					if (newControlScene != null) {
						viewer.removeScene(controlScene);
						controlScene = newControlScene;
						viewer.addScene(newControlScene);
					}
					templateNameToTemplate = (HashMap) in.get("templateNameToTemplate");
					templateToScene = (HashMap) in.get("templateToScene");
					templateToCount.clear();
					for (Object templateName : templateNameToTemplate.keySet()) {
						addTemplate(
								(ModellingTool)templateNameToTemplate.get(templateName),
								templateToScene.get(templateName)
						);
					}
				}
				persistencePeer.restorePersistentProperties(in);
				HashMap map =
					(HashMap) in.get("modelTools");
				for (Object t : map.keySet()) {
					addTool((ModellingTool)t, (String)map.get(t));
				}
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			}
		}
	}

	private static XStream createXStream() {
		XStream xstream = new XStream(new PureJavaReflectionProvider());
		xstream.registerConverter(new FontConverter(xstream.getClassMapper()));
		return xstream;
	}

	public void save() {
		//JFileChooser chooser = ModellerIO.createFileChooser(null);
		File f = persistencePeer.saveFileDialog();
		XStream xstream = createXStream();
		FileWriter fileWriter;
		HashMap<String, Object> out = new HashMap<String, Object>();
		persistencePeer.storePersistentProperties(out);
		out.put("controlScene", controlScene);
		out.put("viewScene", viewScene);
		out.put("templateNameToTemplate", templateNameToTemplate);
		out.put("templateToScene",templateToScene);
		out.put("modelTools", toolToTemplate);
		try {
			fileWriter = new FileWriter(f);
			
			xstream.toXML(out, fileWriter);
			fileWriter.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Adds an object to the list of inspectable objects in the inspect menu.
	 * 
	 * @param o The object to be inspected.
	 * @param name The name of the object to be used in the inspect menu.
	 * @param excludedProperties Contains the names of properties that should not be inspected.
	 */
	public void addObjectToInspect(
			final Object o,
			String name
	) {
		inspectMenu.add(
				new ActionTree(
						new AbstractAction(name) {
							final static long serialVersionUID = 1984;
							
							public void actionPerformed(ActionEvent e) {
								inspector.inspect(
										o,
										objectToExcludedProperties.get(o)
								);
							}
						}
				)
		);
	}
	
	public Object getSceneForTemplate(String templateName) {
		return templateToScene.get(templateName);
	}
	
	public Set<String> getExcludedProperties(Object o) {
		return objectToExcludedProperties.get(o);
	}
	
	public void setExcludedProperties(Object o, Set<String> properties) {
		objectToExcludedProperties.put(o,properties);
	}
	
	/**
	 * Tells all modelling tools to store the state of their models.
	 * 
	 * @see ModellingTool#storeModelState()
	 */
	public void storeModelStates() {
		for (ModellingTool tool : tools)  {
			tool.storeModelState();
		}
	}
	
	/**
	 * Tells all modelling tools to recall the stored state of their model.
	 * 
	 * @see ModellingTool#recallModellState()
	 */
	public void recallModelStates() {
		updatingViewFromEnvironment = false;
		for (ModellingTool tool : tools)  {
			tool.recallModelState();
		}
		updatingViewFromEnvironment = true;
		viewer.render();
	}
	
	/**
	 * Tells all modelling tools that they should update their state from their model
	 * because something might have happened to the model (presumably as a result of
	 * interaction with the model container).
	 */
	public void updateTools() {
		for (ModellingTool tool : tools) {
			tool.stateChanged(new ChangeEvent(tool.getModel()));
		}
		viewer.render();
	}
	
	private void paste() {
		addTool(
				clipBoardTool.clone(),
				clipBoardTemplate,
				viewer.getLastMenuLocation()
		);
	}
	
	private void newTool(String templateName) {
		ModellingTool template = templateNameToTemplate.get(templateName);
		addTool(
				template.clone(),
				template.getName(),
				viewer.getLastMenuLocation()
		);
	}
	
	private void addTool(
			final ModellingTool tool,
			final String template,
			double[] location
	) {
		// cook up a suitable name for the new tool
		Integer id = ((Integer)templateToCount.get(template));
		tool.setName(template+" "+id.toString());
		templateToCount.put(template, new Integer(id.intValue()+1));
		
		// send the tool to its location
		if (location != null) tool.moveTo(location);
		addTool(tool, template);
	}
	
	public void addTool(
			final ModellingTool tool,
			final String template
	) {
		// we want to forward change events from the inspector
		modelToTool.put(tool.getModel(),tool);
		
		toolToTemplate.put(tool, template);
		
		//start listening to model changes
		tool.addChangeListener(toolListener);
		
		// prepare the context menu for the tool
		ActionTree toolMenu = new ActionTree();
		ActionTree toolInspectMenu = new ActionTree("inspect");
		toolInspectMenu.add(new AbstractAction("model") {
			final static long serialVersionUID = 1984;
			public void actionPerformed(ActionEvent e) {
				inspector.inspect(tool.getModel(),null);
			}
		});
        toolInspectMenu.add(new AbstractAction("controller") {
        	final static long serialVersionUID = 1984;
        	public void actionPerformed(ActionEvent e) {
        		String templateName = toolToTemplate.get(tool);
        		ModellingTool t = templateNameToTemplate.get(templateName);
        		Set<String> excluded = new HashSet<String>();
        		Set<String> alreadyExcluded = objectToExcludedProperties.get(t);
        		if (alreadyExcluded != null) {
        			excluded.addAll(objectToExcludedProperties.get(t));
        		}
        		excluded.add("modelType");
        		inspector.inspect(tool,excluded);
        	}
        });
        toolInspectMenu.add(new AbstractAction("control scene") {
        	final static long serialVersionUID = 1984;
        	public void actionPerformed(ActionEvent e) {
        		Set<String> excluded = getExcludedProperties(tool.getControlScene());
        		if (excluded == null) excluded = getExcludedProperties(getControlScene());
        		inspector.inspect(
        				tool.getControlScene(),
        				excluded
        		);
        	}
        });
        toolInspectMenu.add(new AbstractAction("view scene") {
        	final static long serialVersionUID = 1984;
        	public void actionPerformed(ActionEvent e) {
        		boolean defined = objectToExcludedProperties.containsKey(tool.getViewScene());
        		Set<String> excluded = defined ? getExcludedProperties(tool.getViewScene()) :
        			objectToExcludedProperties.get(viewScene);
        		inspector.inspect(
        				tool.getViewScene(),
        				excluded
        		);
        	}
        });
		toolMenu.add(toolInspectMenu);
        toolMenu.add(new AbstractAction("store model state") {
        	final static long serialVersionUID = 1984;
        	public void actionPerformed(ActionEvent e) {
        		tool.storeModelState();
        	}
        });
        toolMenu.add(new AbstractAction("recall model state") {
        	final static long serialVersionUID = 1984;
        	public void actionPerformed(ActionEvent e) {
        		tool.recallModelState();
        	}
        });
        if (copyEnabled) {
        	toolMenu.add(new AbstractAction("copy") {
        		final static long serialVersionUID = 1984;
        		public void actionPerformed(ActionEvent e) {
        			copyTool(tool, template);
        		}
        	});
        	toolMenu.add(new AbstractAction("delete") {
        		final static long serialVersionUID = 1984;
        		public void actionPerformed(ActionEvent e) {
        			removeTool(tool,template);
        		}
        	});
        	toolMenu.add(new AbstractAction("use as template") {
        		final static long serialVersionUID = 1984;
        		public void actionPerformed(ActionEvent e) {
        			setTemplate(template,tool);
        		}
        	});
        }
        int n = toolMenu.getItemCount();
        for (int j=0; j<n; j++) {
        	tool.addActionTree(toolMenu.getItem(j));
        }
        
        // add scenes to their parents
        Object controller = tool.getControlScene();
		if (controller != null) {
			viewer.addChildSceneToParentScene(
					controller,
					controlScene
			);
		}
		Object view = tool.getViewScene();
		Object parent = template == null ? viewScene : templateToScene.get(template);
		if (view != null) {
			viewer.addChildSceneToParentScene(
					view,
					parent
			);
		};
		
		// add tool to the list of tools for broadcasting storeModelState(),
		// recallModelState() and updateFromModel()
		tools.add(tool);
		
		// add model to container
		modelContainer.addModel(tool.getModel(),tool.getModelType());
	}
	
	/**
	 * Adds a type of modelling tool to be included in the "new" menu. Throws an
	 * IllegalStateException if called after <code>finalize()</code>.
	 * 
	 * @param template A modelling tool serving as a template for creating new
	 * modelling tools by copying the template. 
	 * @param scene The scene to which all view scenes of copies of this template will
	 * be added as childeren (so they can inherit their properties).
	 * @param excludedPropertiesForScene Contains the names of those properties
	 * that should not be exposed when inspecting <code>scene</code>.
	 */
	public void addTemplate(
			ModellingTool template,
			final Object scene
	) {
		final String templateName = template.getName();
		
		// initialize counting of copies of this template
		templateToCount.put(templateName,new Integer(0));
		
		// add items to menues
		newMenu.add(new AbstractAction(template.getName()) {
			final static long serialVersionUID = 1984;
			public void actionPerformed(ActionEvent e) {
				newTool(templateName);
			}
		});
		sceneInspectMenu.add(new AbstractAction(template.getName()) {
			final static long serialVersionUID = 1984;
			public void actionPerformed(ActionEvent e) {
        		inspector.inspect(scene, objectToExcludedProperties.get(scene));
        	}
		});
		
		templateNameToTemplate.put(template.getName(), template);
		// view scenes of tools will be added to this scene
		templateToScene.put(template.getName(), scene);
		
		// add the scene to the viewer
		viewer.addChildSceneToParentScene(scene, viewScene);
	}
	
	public void removeTool(ModellingTool tool) {
		removeTool(tool,toolToTemplate.get(tool));
	}
	
	public void removeTool(ModellingTool tool, String template) {
		
		// stop listening to change events from the tool
		tool.removeChangeListener(toolListener);
		
		toolToTemplate.remove(tool);
		// delete information about the model
		Object model = tool.getModel();
		modelToTool.remove(model);
		modelContainer.removeModel(model,tool.getModelType());
		
		// remove scenes from the scene graph
		Object view = tool.getViewScene();
		Object s = templateToScene.get(template);
		if (s == null) s = viewScene;
		viewer.removeChildSceneFromParentScene(view,s);
		Object control = tool.getControlScene();
		viewer.removeChildSceneFromParentScene(control, controlScene);
		
		// check if some part of the tool is being inspected
		Object insp = inspector.getObject();
		if (insp==view || insp==model || insp == control || insp==tool) {
			inspector.inspect(null,null);
		}
		
		// forget about tool
		tools.remove(tool);;
	}
	
	private void copyTool(ModellingTool tool, String template) {
		if (clipBoardTool == null) {
			pasteAction.setEnabled(true);
		}
		clipBoardTool = tool.clone();
		clipBoardTemplate = template;
	}

	public ActionTree getFileMenu() {
		return fileMenu;
	}

	public ActionTree getEditMenu() {
		return editMenu;
	}

	public ActionTree getInspectMenu() {
		return inspectMenu;
	}

	public Object getControlScene() {
		return controlScene;
	}
	
	public ModelContainer getModelContainer() {
		return modelContainer;
	}
	
	public void setModelContainer(ModelContainer modelContainer) {
		if (this.modelContainer != null) {
			this.modelContainer.removeChangeListener(containerListener);
		}
		if (modelContainer == null) {
			modelContainer = new SimpleModelContainer();
		}
		this.modelContainer = modelContainer;
		modelContainer.addChangeListener(containerListener);
	}

	public Object getViewScene() {
		return viewScene;
	}

	public ModellingViewer getViewer() {
		return viewer;
	}

	public void setViewer(ModellingViewer viewer) {
		this.viewer.removeScene(viewScene);
		this.viewer.removeScene(controlScene);
		viewer.addScene(viewScene);
		viewer.addScene(controlScene);
		this.viewer = viewer;
	}

	public Inspector getInspector() {
		return inspector;
	}

	public void setInspector(Inspector inspector) {
		this.inspector.inspect(null,null);
		this.inspector.removeChangeListener(inspectorListener);
		inspector.addChangeListener(inspectorListener);
		this.inspector = inspector;
	}

	public ActionTree getViewerMenu() {
		return viewerMenu;
	}

	public boolean isCopyEnabled() {
		return copyEnabled;
	}

	public void setCopyEnabled(boolean copyEnabled) {
		this.copyEnabled = copyEnabled;
	}
	
	private void setTemplate(String templateName, ModellingTool tool) {
		ModellingTool newTemplate = tool.clone();
		newTemplate.setName(templateName);
		templateNameToTemplate.put(templateName, newTemplate);
	}

	public ActionTree getNewMenu() {
		return newMenu;
	}

	public HashMap<ModellingTool, String> getToolToTemplate() {
		return toolToTemplate;
	}
}